//----------------------------------------------
// PYRAMID -- Double precision version
//----------------------------------------------
// Copyright (C) 1993, Nicholas Centanni
//----------------------------------------------
#include <graphics.h>
#include <op_view.h>
#include <op_color.h>
#include "d_pyramd.h"

d_point pyramid_vertices[] =
{
  { -BASE_SIZE, 0, -BASE_SIZE },
  {  BASE_SIZE, 0, -BASE_SIZE },
  {  BASE_SIZE, 0,  BASE_SIZE },
  { -BASE_SIZE, 0,  BASE_SIZE },
  {  0, PYR_HEIGHT, 0 },
};

d_surface pyramid_surfaces[] =
{
  { { 4, 1, 0 }, { 0.7, 0.55, 0.5 }, 0.4, 0.6, 1, 3 },
  { { 4, 2, 1 }, { 0.7, 0.55, 0.5 }, 0.4, 0.6, 2, 3 },
  { { 4, 3, 2 }, { 0.7, 0.55, 0.5 }, 0.4, 0.6, 3, 3 },
  { { 4, 0, 3 }, { 0.7, 0.55, 0.5 }, 0.4, 0.6, 4, 3 },
};

//------------------------------------------------------------------------
// OBJECT constructor
//------------------------------------------------------------------------
d_Pyramid::d_Pyramid(d_Environ *env) : d_Object(env)
{
  read_vertices(pyramid_vertices, 5);
  read_surfaces(pyramid_surfaces, 4);
  define_surfaces();
  set_draw_method(SHADE_FACING);
  set_fill_method(SOLID_FILL);
  offset_shades(0);

  // Initialize a bounding box
  st_box[0].x = getmaxx();
  st_box[0].y = getmaxy();
  st_box[1].x = st_box[1].y = 0;
}

//-------------------------------------------------------------------------
// Calculate the bounding box to facilitate a fast screen erase
//-------------------------------------------------------------------------
void d_Pyramid::bounding_box()
{
  int page = graphics->active_page;
  int size = sizeof(i_screen);
  char *p = pixels + (size * page);
  i_screen *pix;

  // Prepare the bounding box structure
  memcpy(&b_box, &st_box, sizeof(b_box));

  // Scan all vertices
  for (int i = 0; i < vert_tot; i++)
  {
    pix = (i_screen *)p;

    // See if this coordinate can update the bounding box
    if (pix->x <= b_box[0].x)
      b_box[0].x = pix->x;
    if (pix->x >= b_box[1].x)
      b_box[1].x = pix->x;
    if (pix->y <= b_box[0].y)
      b_box[0].y = pix->y;
    if (pix->y >= b_box[1].y)
      b_box[1].y = pix->y;

    p += pixel_size;
  }
}

//------------------------------------------------------------------------
// OBJECT drawing methods
//------------------------------------------------------------------------
void d_Pyramid::draw(int state, int d_method, int in_window)
{
  int page = graphics->active_page;
  int i, j, color, size = sizeof(i_screen);
  i_screen *s1, *s2, *pix;
  char *p = pixels + (size * in_window) + (size * page);
  int method = d_method == USE_DEFAULT ? draw_method : d_method;
  int do_shade = OFF;

  if (method == POINTS)
  {
    // Draw all vertices as points
    for (i = 0; i < vert_tot; i++)
    {
      color = state == ON ? WHITE : 0;
      pix = (i_screen *)p;

      if (pix->visible)
        plot_1(pix->x, pix->y, color);
      p += pixel_size;
    }
  }

  else if (method == WIRE_SIDES)
  {
    int st[2], en[2];
    color = state == ON ? WHITE : 0;

    // Draw the visible sides
    for (i = 0; i < surface_tot; i++)
    {
      int edges = surfaces[i].edge_tot;

      setcolor(color);
      s1 = (i_screen *)(p + (pixel_size * surfaces[i].v_index[edges-1]));
      for (j = 0; j < surfaces[i].edge_tot; j++)
      {
        s2 = (i_screen *)(p + (pixel_size * surfaces[i].v_index[j]));
        if (s1->visible && s2->visible)
        {
          st[0] = s1->x;
          st[1] = s1->y;
          en[0] = s2->x;
          en[1] = s2->y;
          n_line(st[0], st[1], en[0], en[1], color, 0);
        }
        s1 = s2;
      }
    }
  }

  else if (method == SHADE_FACING)
  {
    struct palettetype pal;
    getpalette(&pal);

    // Calculate side visibility for the object
    if (state == ON)
      facing_surfaces(in_window);

    // Determine shading requirements
    if (moved || light->changed)
    {
      moved = OFF;
      do_shade = ON;
    }

    if (state == ON)
    {
      // Render all visible sides
      for (i = 0; i < surface_tot; i++)
      {
        d_surface *s = &surfaces[i];

        if (do_shade && s->facing)
        {
          shade_surface(i);
          int r = s->lighting.r * 63;
          int g = s->lighting.g * 63;
          int b = s->lighting.b * 63;
          setrgbpalette(pal.colors[s->shade], r, g, b);
        }
        if (s->facing)
          fill_surface(s, state, in_window);
      }
    }
    else
    {
      bounding_box();
      setfillstyle(SOLID_FILL, 0);
      bar(b_box[0].x, b_box[0].y, b_box[1].x, b_box[1].y);
    }
  }
}

//----------------------------------------------------------------------
// d_Pyramid replicator
//----------------------------------------------------------------------
d_Pyramid *d_Pyramid::clone(d_Environ *env)
{
  d_Pyramid *new_object = NULLP;

  if ((new_object = new d_Pyramid(env)) == NULL)
    bail(MEM_ERR);
  else
    new_object->set_center(center.x, center.y, center.z);

  return new_object;
}

//------------------------------------------------------------------------
// Add an offset to all surface shades
//------------------------------------------------------------------------
void d_Pyramid::offset_shades(int offset)
{
  for (int i = 0; i < surface_tot; i++)
    surfaces[i].shade += offset;
  shade_offset = offset;
}
