//----------------------------------------------
// S_CUBE -- XYZ++ example program
//----------------------------------------------
// Animates a solid-filled cube, using the
// SPLIT/VGA mode.
//----------------------------------------------
// Copyright (C) 1993, Nicholas Centanni
//----------------------------------------------
#include <graphics.h>
#include <dos.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#include <op_graph.h>
#include <op_view.h>

#define VIEW_SCREEN   0
#define UP            1
#define DOWN          2
#define ON            1
#define OFF           0

Graphics graphics;       // Declare a graphics object
Viewer viewer;           // Declare a 3-D viewer object

// Define the cube size
#define SIZE      60
#define SIZE_2   (SIZE * 2)

// Define the cube center coordinates
#define CX        10
#define CY        10
#define CZ        10

// Define a vertex for the center of the cube
l_vertex center = { CX, CY, CZ };

// Define the cube vertex structure
struct cube_point
{
  l_point world;    // World coordinate
  l_point rotated;  // Rotated coordinates
  i_screen pix[2];  // Two sets of screen pixels, one for each video page
};

// Define the cube
struct cube_point cube[] =
{
  { CX - SIZE, CY - SIZE, CZ - SIZE }, // Bottom four vertices
  { CX + SIZE, CY - SIZE, CZ - SIZE },
  { CX + SIZE, CY - SIZE, CZ + SIZE },
  { CX - SIZE, CY - SIZE, CZ + SIZE },

  { CX - SIZE, CY + SIZE, CZ - SIZE },    // Top four vertices
  { CX + SIZE, CY + SIZE, CZ - SIZE },
  { CX + SIZE, CY + SIZE, CZ + SIZE },
  { CX - SIZE, CY + SIZE, CZ + SIZE },
};

// Polygon structure for the use of the
// BGI fillpoly() function
int poly[8];

// The cube_side array will keep track of the vertices
// and surface normals of each of the 6 cube surfaces
struct cube_side
{
  int corner[4];         // The four corners
  int visible[2];
};

// Initialize the cube side descriptions
struct cube_side sides[] =
{
  { 0, 1, 2, 3 },  // Bottom
  { 1, 0, 4, 5 },  // North
  { 1, 5, 6, 2 },  // East
  { 3, 2, 6, 7 },  // South
  { 0, 3, 7, 4 },  // West
  { 7, 6, 5, 4 },  // Top
};

// Supporting functions
void display(int color_state);
void animate(void);

//-------------------------------------------------------------------------
// Program main
//-------------------------------------------------------------------------
void main()
{
  // Start graphics
  registerbgidriver(EGAVGA_driver);
  graphics.on(VGA, VGAHI, "");
  delay(800);  // Give auto-sync screens a chance to settle down

  // Define an animation-capable viewing window
  graphics.define_port(VIEW_SCREEN, 0, 0, getmaxx(), SPLIT_ROW - 1,
                       PAGING_ON, SPLIT_ON);

  // Set up the split screen
  graphics.enable_split_screen();
  graphics.paging_on(OFS_ADR, PR_PAGING);
  graphics.clear_screen();

  // The 3-D zoom setting MUST take place AFTER
  // the above calls to define_port and set_port
  graphics.set_port(VIEW_SCREEN);
  viewer.set_zoom(90);

  // Run the animation until a key is pressed
  animate();

  // Return to text mode
  graphics.off();
}

//-------------------------------------------------------------------------
// Move past the object in space
//-------------------------------------------------------------------------
void animate()
{
  int rotate_chg = 20;

  // Place the viewer at a position in space: x, y, z
  viewer.set_position(700, 520, 300);

  // Set the center of interest to be the middle of the cube
  viewer.set_center(center.pos.x, center.pos.y + 30, center.pos.z);
  viewer.view_center();  // Aim the viewer at the center
  graphics.reset_pages();

  // Animate
  while (!kbhit())
  {
    // Erase the object
    graphics.chg_active_page();
    display(OFF);

    // Change the viewer position
    viewer.orbit(rotate_chg);
    viewer.chg_heading(rotate_chg);

    // Redisplay the object from the new angle
    display(ON);
    graphics.chg_visual_page();

    delay(10);
  }

  // Read the exit keystroke so it
  // doesn't pass on to DOS
  getch();
}

//-------------------------------------------------------------------------
// Display the cube at its current position,
// from the current point of view.
//-------------------------------------------------------------------------
void display(int color_state)
{
  int i, j;
  long v1, v2, w1, w2;
  int w = graphics.active_page;
  int y_off = w ? 0 : LO_LINES;
  int colors[6] =
   { GREEN, BLUE, MAGENTA, LIGHTGREEN, LIGHTBLUE, LIGHTMAGENTA };

  // Only calculate coordinates when drawing the visible image
  if (color_state == ON)
  {
    // Calculate the viewing matrix for this angle of view
    viewer.new_view();

    // Calculate the 3-D location of each cube vertex
    for (i = 0; i < 8; i++)
      viewer.translate(&cube[i].world, &cube[i].pix[w]);

    // Determine the visibility of each cube side
    for (i = 0; i < 6; i++)
    {
      struct cube_side *s = &sides[i];
      i_screen *px0 = &cube[s->corner[0]].pix[w];
      i_screen *px1 = &cube[s->corner[1]].pix[w];
      i_screen *px2 = &cube[s->corner[2]].pix[w];

      // Determine the orientation of the face
      v1 = px1->x - px0->x;
      v2 = px1->y - px0->y;
      w1 = px2->x - px1->x;
      w2 = px2->y - px1->y;

      // Store the visibility flag
      sides[i].visible[w] = ((v1 * w2 - v2 * w1) < 0);
    }
    setcolor(WHITE);
  }
  else
    setcolor(BLACK);

  // Draw the visible cube sides
  for (i = 0; i < 6; i++)
  {
    if (sides[i].visible[w])
    {
      struct cube_side *s = &sides[i];

      // Draw the side outlines
      moveto(cube[s->corner[3]].pix[w].x,
             cube[s->corner[3]].pix[w].y + y_off);
      for (j = 0; j < 4; j++)
        lineto(cube[s->corner[j]].pix[w].x,
               cube[s->corner[j]].pix[w].y + y_off);

      // Build the polygon description
      for (j = 0; j < 4; j++)
        memcpy(&poly[j * 2], &cube[s->corner[j]].pix[w], 4);
      for (j = 1; j < 8; j += 2)
        poly[j] += y_off;

      // Set the current fill color
      if (color_state == ON)
        setfillstyle(SOLID_FILL, colors[i]);
      else
        setfillstyle(SOLID_FILL, BLACK);

      // Fill the side
      fillpoly(4, poly);
    }
  }
}

