//--------------------------------------------------------
// Defines a generic 3-D object
//--------------------------------------------------------
#ifndef __SCENE_H
#define __SCENE_H
#ifndef __STRING_H
#include <string.h>
#endif
#ifndef __GRAPHICS_H
#include <graphics.h>
#endif
#ifndef __OP_VIEW_H
#include <op_view.h>
#endif
#ifndef __OP_GRAPH_H
#include <op_graph.h>
#endif
#ifndef __OP_LIGHT_H
#include <op_light.h>
#endif

// Drawing modes
#define NOT_SET         -1
#define USE_DEFAULT      0
#define POINTS           1
#define WIRE_SIDES       2
#define WIRE_LINES       3
#define SOLID_FRAME      4
#define SHADE_FACING     5
#define SHADE_Z_SORTED   6

// Filling methods
#define BGI_FILL         0
#define DITHER_FILL      1
#define SOLID_FILL       2

// Graphics modes
#define EGA_16          -1
#define VGA_16          -2
#define VGA_SPLIT_16    -3

// Full screen port constant
#define FULL_SCREEN      0

//-------------------------------------------------------------
// Double precision Classes -----------------------------------
//-------------------------------------------------------------
//-------------------------------------------------------------
// Environment class: Contains information about the program
//-------------------------------------------------------------
class d_Environ
{
 public:
   Viewer *viewer;         // Connection to the 3-D viewer class
   Graphics *graphics;     // Connection to the graphics class
   Vect *vect;             // Vector geometry class
   d_Light *light;         // Connection to shading class
   int pixel_size;         // The size of one pixel array element

   d_Environ(Viewer *v, Graphics *g, Vect *vec, d_Light *l = NULLP,
             int max_pages = 1, int max_views = 1);
   ~d_Environ(void) {}
};

//-------------------------------------------------------------
// Generic 3-D object description class
//-------------------------------------------------------------
class d_Object
{
 public:
   Viewer *viewer;         // Connection to the 3-D viewer class
   Graphics *graphics;     // Connection to a graphics class
   Vect *vect;             // Vector geometry class
   d_Light *light;         // Connection to the shading class
   d_point *vertices;      // Vertex array
   int vert_tot;           // Total object vertices
   d_surface *surfaces;    // Surface array
   int surface_tot;        // Total object surfaces
   LINE *lines;            // Line array
   int line_tot;           // Total object lines
   char *pixels;           // Pixel array
   int pixel_size;         // The size of one pixel array element
   int draw_method;        // Current drawing method
   int fill_method;        // Default surface fill method
   double orbit_radius;    // Pre-set radius for animation orbits
   int orbit_pt;           // Pre-set vertex for animation orbits
   d_point *orbit_center;  // Pre-set orbit pivot-point
   d_point center;         // Object center point for SCENE sorting
   double center_z;        // Object center Z value for SCENE sorting
   int moved;              // Set when the object has moved
   int force_shading;      // Set to override shade-skipping logic

   // Default constructor and destructor
   d_Object(void);
   d_Object(d_Environ *env);
   ~d_Object(void);

   // Clear existing arrays
   void delete_vertices(void);
   void delete_surfaces(void);
   void delete_lines(void);

   // Fill the arrays from pre-defined lists (OPTIONAL)
   void read_vertices(d_point *vert_source, int v_tot);
   void read_surfaces(d_surface *surface_source, int s_tot);
   void read_lines(LINE *line_source, int l_tot);

   // Object Point management
   void set_vertex(int index, d_point *pt);
   void copy_vertex(int v_index, d_point *dest);
   void set_center(double x, double y, double z)
     { center.x = x; center.y = y; center.z = z; }

   // Move Object
   void move(double x, double y, double z);

   // Object surface management
   void facing_surfaces(int in_window);
   void fill_surface(d_surface *surface, int state, int in_window);
   void define_surface(d_plane *p, d_point *base, d_point *v1, d_point *v2);
   void define_surfaces(void);
   void shade_surface(int surface_index);
   void force_shade(void) { force_shading = ON; }

   // Vertex translate calculation
   void translate(int in_window);

   // Object DRAW functions
   void set_draw_method(int method) { draw_method = method; }
   void set_fill_method(int method) { fill_method = method; }
   virtual void draw(int state, int method = USE_DEFAULT,
                     int in_window = FULL_SCREEN) = 0;

   // Object orbit management
   void define_orbit(int pt_index, d_point *center);
   void orbit(int heading);
};

//-------------------------------------------------------------
// Generic Scene description class
//-------------------------------------------------------------
class d_Scene
{
 private:
   int max_objects;        // Used to size the object array

 public:
   Viewer *viewer;         // Connection to the 3-D viewer class
   Graphics *graphics;     // Connection to a graphics class
   d_Light *light;         // Connection to shading class
   d_Object **objects;     // CLASS Array of scene object pointers
   int tot_objects;        // Total of objects currently defined

   // Scene constructor and destructor
   d_Scene(d_Environ *env, int max_obj);
   ~d_Scene(void);

   // Lighting
   void set_light_source(int which, double x, double y, double z,
	                      double r, double g, double b)
     { light->define_source(which, x, y, z, r, g, b); }
   void set_light_source(int which, d_source *lt)
     { light->define_source(which, lt->world.x, lt->world.y, lt->world.z,
                            lt->color.r, lt->color.g, lt->color.b); }

   // External graphics control
   void start_graphics(int gr_mode);
   void end_graphics(void);

   // Object management
   void add_object(d_Object *new_object) 
     { objects[tot_objects++] = new_object; }
   void translate(int in_window = 0);
   void draw(int state, int method = USE_DEFAULT, int in_window = 0);
   void sort_objects(void);
};

//-------------------------------------------------------------
// Fixed Point Classes ----------------------------------------
//-------------------------------------------------------------
//-------------------------------------------------------------
// Environment class: Contains information about the program
//-------------------------------------------------------------
class f_Environ
{
 public:
   Viewer *viewer;         // Connection to the 3-D viewer class
   Graphics *graphics;     // Connection to the graphics class
   Vect *vect;             // Private Vector geometry class
   f_Light *light;         // Connection to shading class
   int pixel_size;         // The size of one pixel array element

   f_Environ(Viewer *v, Graphics *g, Vect *vec, f_Light *l = NULLP,
             int max_pages = 1, int max_views = 1);
   ~f_Environ(void) {}
};

//-------------------------------------------------------------
// Generic 3-D object description class
//-------------------------------------------------------------
class f_Object
{
 public:
   Viewer *viewer;         // Connection to the 3-D viewer class
   Graphics *graphics;     // Connection to a graphics class
   Vect *vect;             // Private Vector geometry class
   f_Light *light;         // Connection to the shading class
   f_point *vertices;      // Vertex array
   int vert_tot;           // Total object vertices
   f_surface *surfaces;    // Surface array
   int surface_tot;        // Total object surfaces
   LINE *lines;            // Line array
   int line_tot;           // Total object lines
   char *pixels;           // Pixel array
   int pixel_size;         // The size of one pixel array element
   int draw_method;        // Current drawing method
   int fill_method;        // Default surface fill method
   fixed orbit_radius;     // Pre-set radius for animation orbits
   int orbit_pt;           // Pre-set vertex for animation orbits
   f_point *orbit_center;  // Pre-set orbit pivot-point
   f_point center;         // Object center point for SCENE sorting
   fixed center_z;         // Object center Z value for SCENE sorting
   int moved;              // Set when the object has moved
   int force_shading;      // Set to override shade-skipping logic

   // Default constructor and destructor
   f_Object(void);
   f_Object(f_Environ *env);
   ~f_Object(void);

   // Clear existing arrays
   void delete_vertices(void);
   void delete_surfaces(void);
   void delete_lines(void);

   // Fill the arrays from pre-defined lists (OPTIONAL)
   void read_vertices(f_point *vert_source, int v_tot);
   void read_surfaces(f_surface *surface_source, int s_tot);
   void read_lines(LINE *line_source, int l_tot);

   // Object Point management
   void set_vertex(int index, f_point *pt);
   void copy_vertex(int v_index, f_point *dest);
   void set_center(fixed x, fixed y, fixed z)
     { center.x = x; center.y = y; center.z = z; }

   // Move Object
   void move(fixed x, fixed y, fixed z);

   // Object surface management
   void facing_surfaces(int in_window);
   void fill_surface(f_surface *surface, int state, int in_window);
   void define_surface(f_plane *p, f_point *base, f_point *v1, f_point *v2);
   void define_surfaces(void);
   void shade_surface(int surface_index);
   void force_shade(void) { force_shading = ON; }

   // Vertex translate calculation
   void translate(int in_window);

   // Object DRAW functions
   void set_draw_method(int method) { draw_method = method; }
   void set_fill_method(int method) { fill_method = method; }
   virtual void draw(int state, int method = USE_DEFAULT,
                     int in_window = FULL_SCREEN) = 0;

   // Object orbit management
   void define_orbit(int pt_index, f_point *center);
   void orbit(int heading);
};

//-------------------------------------------------------------
// Generic Scene description class
//-------------------------------------------------------------
class f_Scene
{
 private:
   int max_objects;        // Used to size the object array

 public:
   Viewer *viewer;         // Connection to the 3-D viewer class
   Graphics *graphics;     // Connection to a graphics class
   f_Light *light;         // Connection to shading class
   f_Object **objects;     // CLASS Array of scene objects
   int tot_objects;        // Total of objects currently defined

   // Scene constructor and destructor
   f_Scene(f_Environ *env, int max_obj);
   ~f_Scene(void);

   // Lighting
   void set_light_source(int which, fixed x, fixed y, fixed z,
	                      fixed r, fixed g, fixed b)
     { light->define_source(which, x, y, z, r, g, b); }
   void set_light_source(int which, f_source *lt)
     { light->define_source(which, lt->world.x, lt->world.y, lt->world.z,
                            lt->color.r, lt->color.g, lt->color.b); }

   // External graphics control
   void start_graphics(int gr_mode);
   void end_graphics(void);

   // Object management
   void add_object(f_Object *new_object) 
     { objects[tot_objects++] = new_object; }
   void translate(int in_window = 0);
   void draw(int state, int method = USE_DEFAULT, int in_window = 0);
   void orbit(int obj_index, int heading);
   void sort_objects(void);
};
#endif

