//----------------------------------------------------------------
// XYZ++ Basic Types and Definitions
//----------------------------------------------------------------
#ifndef __OP_TYPES_H
#define __OP_TYPES_H
#ifndef __OPUS_H
#include <opus.h>
#endif

//-------------------------------------------------------------
// Fixed Point Definitions
//-------------------------------------------------------------
typedef long fixed;     // 16.16 bit fixed point data type
#define I2F(x) (((long)(int)x) << 16)
#define D2F(x) ((long) (x * 65536.0 + 0.5))
#define F2I(f) ((int)(f >> 16))
#define F2D(f) (((double)(f & 0xffff) / 65536.0) + (int)(f >> 16))
#define FX_0 0L
#define FX_1 65536L
extern "C" fixed f_mul(fixed, fixed);
extern "C" fixed f_div(fixed, fixed);
extern "C" fixed fixed_sqrt(fixed value);
extern fixed f_sqrt(double value);
extern fixed cos_table[];

//----------------------
// Basic Elements
//----------------------
struct i_point          // 3D coordinate described
{                       // with integer components
  long x;
  long y;
  long z;
};

struct l_point          // 3D coordinate described
{                       // with long components
  long x;
  long y;
  long z;
};

struct d_point          // 3D coordinate described
{                       // with double components
  double x;
  double y;              
  double z;              
};
                        
struct f_point          // 3D coordinate described 
{                       // with fixed point components
  fixed x;               
  fixed y;               
  fixed z;
};

struct s_point          // 2D coordinate 
{                       
  int x;                // Horizontal
  int y;                // Vertical
};

struct i_screen         // 2D screen coordinate with visibility flag
{                       
  int x;                // Horizontal direction 
  int y;                // Vertical direction 
  int visible;          // Screen visibility flag
};

struct l_vertex         // Object vertex - long version
{                       
  l_point pos;          // 3D coordinate
  i_screen pixel;       // 2D screen coordinate
  int origin;           // 3D clipping flag
};

struct d_vertex         // Object vertex - double version
{                       
  d_point pos;          // 3D coordinate
  i_screen pixel;       // 2D screen coordinate
  int origin;           // 3D clipping flag
};

struct f_vertex         // Object vertex - fixed point version
{                       
  f_point pos;          // 3D coordinate
  i_screen pixel;       // 2D screen coordinate
  int origin;           // 3D clipping flag
};

// Line description
typedef struct
{
  int st;               // Index to world coordinates of START point
  int en;               // Index to world coordinates of END point
} LINE;

struct d_plane          // Plane description - double version
{      
  d_point base;         // Base point end of the surface normal
  d_point normal;       // Surface normal vector
  double scalar;        // Scalar component of the plane equation
};

struct f_plane          // Plane description - fixed version
{      
  f_point base;         // Base point end of the surface normal
  f_point normal;       // Surface normal vector
  fixed scalar;         // Scalar component of the plane equation
};

// Derived forms
typedef d_point d_rect[4];
typedef d_point d_line[2];
typedef d_vertex d_poly[8];
typedef f_point f_rect[4];
typedef f_point f_line[2];
typedef f_vertex f_poly[8];

//-----------------------
// Color definitions
//-----------------------
struct d_color
{ 
  double r;      // Red Component
  double g;      // Green Component
  double b;      // Blue Component
};

struct f_color
{ 
  fixed r;       // Red Component
  fixed g;       // Green Component
  fixed b;       // Blue Component
};

//-----------------------
// Surface definitions
//-----------------------
#define MAX_VERT   4

struct d_surface
{
  int v_index[MAX_VERT];   // Array of world coordinate vertex indices
  d_color color;           // Actual surface color expressed as RGB
  float ambient;           // Percent of ambient light that surface reflects
  float diffuse;           // Percent of diffuse light that surface reflects
  int shade;               // Surface shade intensity for current lighting
  unsigned char edge_tot;  // Number of surface edges
  unsigned char facing;    // Surface visibility flag
  d_color lighting;        // Surface color based on current lighting
  d_plane plane;           // The plane equation for this surface
};

struct f_surface
{
  int v_index[MAX_VERT];   // Array of world coordinate vertex indices
  f_color color;           // Actual surface color expressed as RGB
  float ambient;           // Percent of ambient light that surface reflects
  float diffuse;           // Percent of diffuse light that surface reflects
  int shade;               // Surface shade intensity for current lighting
  unsigned char edge_tot;  // Number of surface edges
  unsigned char facing;    // Surface visibility flag
  f_color lighting;        // Surface color based on current lighting
  f_plane plane;           // The plane equation for this surface
};

//-------------------------
// Lighting definitions
//-------------------------
#define LIGHTSCALE  200   
#define MAX_LIGHTS  5

struct d_source
{ 
  d_point world;  // Location of the light source
  d_color color;  // Color of the emitted light
};

struct f_source
{ 
  f_point world;  // Location of the light source
  f_color color;  // Color of the emitted light
};
#endif
